function [p_new,a_new,b_new,t_new,alpha,scale_a,beta,scale_b,err] = chmm_fbp(net, obss, scale_in)

% [p_new, a_new, b_new, t_new, alpha, scale_a, beta, scale_b,err] = chmm_fbp(net, obs, scale_in)
% This function calculates the forward variables, backward varibles and new parameter
% values based on partials computed using Baum-Welch algorithm.
% INPUTS:
%    net - chmm model
%    obs - observation sample(s)
%    scale_in - input scaling coefficients (optional)
% OUTPUTS:
%    p_new - new pi's (reestimations)
%    a_new - new a_ij's (reestimations)
%    b_new - new b_jk's (reestimations)
%    t_new - new theta's (reestimations)
%    alpha - forward variables
%    scale_a - scaling coefficients associated with alpha's
%    beta - backward variables
%    scale_b - scaling coefficients associated with beta's
%    err - err is 1 if singularity problem occurs, 0 otherwise

err = 0;
ss = net.slice_size;
nh = ss/2;
ns = net.node_sizes(:);
[nobs,T,nsamp] = size(obss);
assert(nobs==nh, '# of samples and # of models do not match');

if nargin == 2
   scale_a = ones(1,T);
   scale_b = ones(1,T);
else
   scale_a = scale_in;
   scale_b = scale_in;
end

% temporary variables
ns1 = ns(1);
no = ns(1+nh);
ns2 = ns1 * ns1;
nh2 = nh * nh;
nso = ns1 * no;
nsh = ns1 * nh;
nsoh = nso * nh;
nsh2 = nsh * nsh;

% complete permutations of alpha or beta indices
n1 = cumprod([1 ns(1:nh)']); 
np = n1(nh+1);
n2 = np ./ n1(2:nh+1); 
pi = zeros(np,nh);
for i = 1 : nh 
   a = repmat(1:ns(i), n1(i), 1); 
   b = a(:); 
   pi(:,i) = repmat(b,n2(i),1); 
end
ip = pi';

% preallocate space for some intermediate variables
aa = zeros(nh);
cc = (1:nh)';
hh = 1:ns1;
nn = 1:np;
pp = zeros(np,np,nh);
pb = zeros(np,nh);
B = zeros(np);
B2 = zeros(np);

% preparation for dt
c1 = repmat(cc',[np*np,1]);
c1 = repmat(c1(:),[nh,1]);
c2 = repmat(cc',[np*np*nh,1]);
c2 = c2(:);
ic1 = repmat(nn',[np*nh2,1]);
ic1 = ic1 + (c1-1)*np;
ic1 = pi(ic1);
jc2 = repmat(nn,[np,1]);
jc2 = repmat(jc2(:),[nh2,1]);
jc2 = jc2 + (c2-1)*np;
jc2 = pi(jc2);
ic1 = ic1 + (jc2-1)*ns1 + (c1-1)*ns2 + (c2-1)*ns2*nh;
ath = reshape(net.aij(ic1),[np,np*nh2]);

% preparation for da
c1 = repmat(cc',[ns1,1]);
c1 = repmat(c1(:),[nh,np]);
c2 = repmat(cc',[nsh,1]);
c2 = repmat(c2(:),[1,np]);
ic1 = repmat(hh',[nh2,np]);
jc2 = c2 + repmat((nn-1)*nh,[nsh*nh,1]);
jc2 = ip(jc2);
iaij = ic1 + (jc2-1)*ns1 + (c1-1)*ns2 + (c2-1)*ns2*nh;
iaij = iaij + repmat((nn-1)*nsh2,[nsh*nh,1]);
th(1,:,:) = net.theta;
th = repmat(th,[ns1,1,1,np]);
th = reshape(th,[nsh*nh,np]);

% indices for b_jk's
ibjk1 = ip + repmat((cc-1)*nso,[1,np]);
ibjk2 = ip + repmat((cc-1)*ns1*T,[1,np]);

% B2
ia1 = repmat(pi,[np,1]);
ia3 = (cc'-1)*ns2;
for c = 1 : nh
   thc = net.theta(:,c);
   ia4 = ia3 + (c-1)*ns2*nh;
   ia2 = repmat((ip(c,:)-1)*ns1,[np,1,nh]);
   ia2 = reshape(ia2,[np*np,nh]);
   ia = repmat(ia4,[np*np,1]);
   ia = ia + ia1 + ia2;
   ia2 = net.aij(ia);
   pp(:,:,c) = reshape(ia2*thc, [np,np]);
end
B2 = prod(pp,3);

% preparation needed for da and dt
B2ct = zeros(np,np,1,nh);
B2ca = zeros(np,nh,np);
for c = 1 : nh
   tmp = pp;
   tmp(:,:,c) = 1;
   tmp1 = prod(tmp,3);
   B2ct(:,:,1,c) = tmp1;
   B2ca(:,c,:) = tmp1;
end
B2ct = repmat(B2ct,[1,1,nh,1]);
B2ct = reshape(B2ct,[np,np*nh2]);
ath = B2ct .* ath;
B2ca = reshape(B2ca,[np,nh*np]);
ii(1,:,:) = ip;
ii = repmat(ii,[ns1,1,1]);
ii = (ii==repmat(hh',[1,nh,np]));
ii = reshape(ii,[nsh,np]);

% release unused space
clear c1 c2 ic1 jc2 ith th pp hh ia ia1 ia2 ia3 ia4 B2ct tmp tmp1;

% reestimated variables
if net.discrete == 1
   b_new = zeros(ns1,no,nh);
   db = zeros(ns1,no,nh);
else
   b_new = zeros(ns1,no,3,nh);
   db = zeros(ns1,no,3,nh);
end
p_new = zeros(ns1,nh);
a_new = zeros(ns1,ns1,nh,nh);
t_new = zeros(nh,nh);
% partials
dp = zeros(ns1,nh);
da = zeros(ns1,ns1,nh,nh);
dt = zeros(nh,nh);
% forward and backward variables
alpha = zeros(np,T,nsamp);
beta = ones(np,T,nsamp);

for n = 1 : nsamp %%% outmost loop w.r.t. number of samples/observations
   obs = obss(:,:,n);
   
   %% partials for each observation
   daij = zeros(nsh2,np);
   dth = zeros(nh2,np);
   bjot = zeros(ns1,no,T,nh);
   bjt = zeros(ns1,T,nh);
   
   %%%%% extended forward algorithm --> begin %%%%%
   % initialization
   ic = ip + repmat((cc-1)*ns1,[1,np]);
   pp1 = net.pi(ic); %%% pi's
   pp1a = prod(pp1);
   
   if net.discrete == 1
      ic = ibjk1 + repmat((obs(:,1)-1)*ns1,[1,np]);
      pb1 = net.bjk(ic); %%% bjk's
   else % continuous observations
      for c = 1 : nh
         bjot(:,:,1,c) = cobs_prob(net.bjo(:,:,:,c),obs(c,1));
         bjt(:,1,c) = sum(bjot(:,:,1,c),2);
      end
      pb1 = bjt(ibjk2);
   end
   pb1a = prod(pb1);
   
   %% alpha
   alpha(:,1,n) = (pp1a .* pb1a)';
   if nargin==2 & n==1
      scale_a(1) = 1/sum(alpha(:,1,n));
   end
   alpha(:,1,n) = alpha(:,1,n) * scale_a(1);
   beta(:,T,n) = beta(:,T,n) * scale_b(T);
   
   % induction
   for t = 2 : T %%% inner loop w.r.t. time
      tb = T + 1 - t; % time index for beta
      
      % B1, B for alpha's and partials
      if net.discrete == 1
         oc = obs(:,t);
         for c = 1 : nh
            a = net.bjk(:,oc(c),c)';
            b = repmat(a,[n1(c),1]);
            pb(:,c) = repmat(b(:),[n2(c),1]);
         end
      else
         for c = 1 : nh
            if t <= tb+1
               bjot(:,:,t,c) = cobs_prob(net.bjo(:,:,:,c),obs(c,t));
               bjt(:,t,c) = sum(bjot(:,:,t,c),2);
            end
            if t < tb+1
               bjot(:,:,tb+1,c) = cobs_prob(net.bjo(:,:,:,c),obs(c,tb+1));
               bjt(:,tb+1,c) = sum(bjot(:,:,tb+1,c),2);
            end
            a = bjt(:,t,c)';
            b = repmat(a,[n1(c),1]);
            pb(:,c) = repmat(b(:),[n2(c),1]);
         end
      end
      B1 = prod(pb,2);
      B = repmat(B1,[1,np]);
      B = B .* B2';
      
      %%% daij %%%
      tmp = repmat(alpha(:,t-1,n),[1,nh*np]);
      tmp = tmp .* B2ca;
      tmp = ii * tmp;
      tmp = reshape(tmp,[nsh*nh,np]);
      tmp = tmp .* repmat(B1',[nsh*nh,1]);
      tmpa = zeros(nsh2,np);
      tmpa(iaij) = tmp;
      daij = daij * B';
      daij = daij + tmpa;
      
      %%% dth %%%
      tmp2 = alpha(:,t-1,n)' * ath;
      tmpt = reshape(tmp2, [np,nh2])';
      tmpt = tmpt .* repmat(B1',[nh2,1]);
      dth = dth * B';
      dth = dth + tmpt;
      
      %%% alpha(t-1) -> alpha(t) %%%
      alpha(:,t,n) = B * alpha(:,t-1,n);
      
      %%% B1, B for beta's
      if net.discrete == 1
         ocb = obs(:,tb+1);
         for c = 1 : nh
            a = net.bjk(:,ocb(c),c)';
            b = repmat(a,[n1(c),1]);
            pb(:,c) = repmat(b(:),[n2(c),1]);
         end
      else
         for c = 1 : nh
            a = bjt(:,tb+1,c)';
            b = repmat(a,[n1(c),1]);
            pb(:,c) = repmat(b(:),[n2(c),1]);
         end
      end
      B1 = prod(pb,2);
      B = repmat(B1,[1,np]);
      B = B .* B2';
      
      %%% beta(tb+1) -> beta(tb)
      beta(:,tb,n) = B' * beta(:,tb+1,n);
      
      %%% scaling %%%
      if nargin==2 & n==1
         scale_a(t) = 1/sum(alpha(:,t,n));
         scale_b(tb) = 1/sum(beta(:,tb,n));
      end
      alpha(:,t,n) = alpha(:,t,n) * scale_a(t);
      beta(:,tb,n) = beta(:,tb,n) * scale_b(tb);
      daij = daij * scale_a(t);
      dth = dth * scale_a(t);
   end
   %%%%% extended forward algorithm <-- end %%%%%
   pn = sum(alpha(:,T,n));
   
   %% dp and db calculated from alpha's and beta's
   if n == 1
      s1 = fliplr([scale_a(2:T) 1]);
      s2 = fliplr(scale_b);
      s1 = s1 ./ s2;
      s2 = cumprod(s1);
      s1 = fliplr(s2);
   end
   x = alpha(:,:,n) .* beta(:,:,n);
   x = x .* repmat(s1,[np,1]);
   gamma1 = zeros(ns1,1,T,nh);
   for c = 1 : nh
      %%% pi %%%
      d = [ns1^(c-1) ns1 ns1^(nh-c)];
      y = reshape(x(:,1),d);
      dp(:,c) = sum(sum(y,3),1)';
      %%% bjk %%%
      u = reshape(x,[d T]);
      v = sum(sum(u,3),1);
      v = reshape(v,[ns1,T]);
      if net.discrete == 1
         for o = 1:no, db(:,o,c) = sum(v(:,obs(c,:)==o),2); end
      else
         w = repmat(sum(v),[ns1,1]);
         gamma1(:,1,:,c) = v ./ w;
      end
   end
   p_new = p_new + dp / pn;
   if net.discrete == 1
      b_new = b_new + db / pn;
   else
      %save bjot bjot bjt;
      gamma = repmat(gamma1,[1,no,1,1]);
      bjt(bjt==0) = 1;
      gamma1(:,1,:,:) = bjt;
      if sum(gamma1(:)==0)>0, err = 1; return; end
      bjot = bjot ./ repmat(gamma1,[1,no,1,1]);
      gamma = gamma .* bjot;
      db(:,:,1,:) = sum(gamma,3); %%% c_jk's
      robs11(1,1,:,:) = obs';
      robs = repmat(robs11,[ns1,no,1,1]);
      db(:,:,2,:) = sum(gamma.*robs,3); %%% u_jk's
      robs = robs - repmat(net.bjo(:,:,2,:),[1,1,T,1]);
      robs = robs .* robs;
      db(:,:,3,:) = sum(gamma.*robs,3); %%% U_jk's
      b_new = b_new + db;
   end
   
   %% partials with multiple observations
   da = da + reshape(sum(daij,2)/pn,[ns1,ns1,nh,nh]);
   dt = dt + reshape(sum(dth,2)/pn,[nh,nh]);
end

% ----- reestimation ----- %
%% pi %%
dp = repmat(sum(p_new),[ns1,1]);
if sum(dp(:)==0)>0, err = 1; return; end
p_new = p_new ./ dp;
%% bjk %%
if net.discrete == 1
   db = repmat(sum(b_new,2),[1,no,1]);
   if sum(db(:)==0)>0, err = 1; return; end
   b_new = b_new ./ db;
else
   db = b_new(:,:,1,:);
   if sum(db(:)==0)>0, err = 1; return; end
   %db(db==0) = 1;
   b_new(:,:,2,:) = b_new(:,:,2,:) ./ db;
   b_new(:,:,3,:) = b_new(:,:,3,:) ./ db;
   db = repmat(sum(b_new(:,:,1,:),2),[1,no,1,1]);
   b_new(:,:,1,:) = b_new(:,:,1,:) ./ db;
end
%% aij %%
a_new = net.aij .* da;
da = repmat(sum(a_new,2),[1,ns1,1,1]);
da(da==0) = 1; % dealing with uncoupled case
a_new = a_new ./ da;
%% theta %%
t_new = net.theta .* dt;
dt = repmat(sum(t_new,1),[nh,1]);
if sum(dt(:)==0)>0, err = 1; return; end
t_new = t_new ./ dt;

return;

