function [dp, da, db, dt, alpha, scale, err] = chmm_partial(net, obss, scale_in)

% [dp, da, db, dt, alpha, scale, err] = chmm_partial(net, obss, scale_in)
%    net - chmm model
%    obss - observation sample(s) (nh x T x nobs, 3-D array)
%    scale_in - input scaling coefficients (optional)
%    dp - partial w.r.t. pi
%    da - partial w.r.t. a_ij
%    db - partial w.r.t. b_jk
%    dt - partial w.r.t. theta
%    alpha - forward variable
%    scale - associated scaling coefficients
%    err - err is 1 if singularity problem occurs, 0 otherwise

err = 0;
ss = net.slice_size;
nh = ss/2;
ns = net.node_sizes(:);
ns1 = ns(1);   % assume all hidden nodes have the same size (ns1)
[nsam,T, nobs] = size(obss);
assert(nsam==nh, '# of samples and # of models do not match');
no = ns(nh+1);

alpha = zeros(ns1,nh,nobs);
dp = zeros(ns1,nh);
da = zeros(ns1,ns1,nh,nh);
if net.discrete == 1
   db = zeros(ns1,no,nh);
else
   db = zeros(ns1,no,3,nh);
   b_new = zeros(ns1,no,3,nh);
end
dt = zeros(nh,nh);

nsh = ns1 * nh;
ns2 = ns1 * ns1;
nh2 = nh * nh;
nsho = nsh * no;
nsh2 = nsh * nsh;
nsh3 = nsh2 * nsh;
hh = (1:ns1)';

i = 1:nh;
i1 = ns2*(i-1)+1;
i2 = ns2*i;
j1 = nh*(i-1)+1;
j2 = nh*i;
a = repmat(1:ns1,[ns1 1]);
ind1 = repmat(a(:),[nh2 1]);
a = repmat(1:nh,[ns2*nh 1]);
ind2 = a(:);
ind = (1:nsh2)';
ind = ind + (ind1-1)*nsh2;
ind = ind + (ind2-1)*nsh2*ns1;

if nargin == 2
   scale = ones(1,T);
else
   scale = scale_in;
end
   
for n = 1 : nobs
   obs = obss(:,:,n);
   
   dpi = zeros(ns1,nh,ns1,nh); % number of pi's ns1 by nh, # of alpha's ns1 by nh
   daij = zeros(ns1,ns1,nh,nh,ns1,nh); % i,j,c1,c2,j,c
   dbjk = zeros(ns1,no,nh,ns1,nh); % j1, k, c1, j, c
   dth = zeros(nh,nh,ns1,nh); % c1, c2, j, c
   bjot = zeros(ns1,no,T,nh);
   bjt = zeros(ns1,T,nh);
   gamma1 = zeros(ns1,1,T,nh);
   
   %%%%% forward algorithm --> begin %%%%%
   for c = 1 : nh %% initialization
      ocpt = net.CPD{c+nh}.CPT; % output node obs matrix
      if net.discrete == 1
         alpha(:,c,n) = net.CPD{c}.prior' .* ocpt(:,obs(c,1));
         dpi(:,c,:,c) = diag(ocpt(:,obs(c,1)));
         dbjk(:,obs(c,1),c,:,c) = diag(net.CPD{c}.prior); % net.pi(:,c)
      else
         bjot(:,:,1,c) = cobs_prob(ocpt,obs(c,1));
         bjt(:,1,c) = sum(bjot(:,:,1,c),2);
         alpha(:,c,n) = net.CPD{c}.prior' .* bjt(:,1,c);
         dpi(:,c,:,c) = diag(bjt(:,1,c));
      end
   end
   if nargin==2 & n ==1
      scale(1) = 1/sum(alpha(:,1,n));
   end
   alpha(:,:,n) = alpha(:,:,n) * scale(1);
   as = repmat(sum(alpha(:,:,n),1),[ns1,1]);
   gamma1(:,1,1,:) = alpha(:,:,n) ./ as;
   dpi = dpi * scale(1);
   dbjk = dbjk * scale(1);
   
   %% temporary variables %%
   tmp = zeros(ns1,nh);
   tmpi = zeros(nsh*ns1,nh);
   tmpa = zeros(nsh2*ns1,nh);
   tmpb = zeros(nsho*ns1,nh);
   tmpt = zeros(nsh*nh,nh);
   ijc = zeros(ns1,ns1,nh);
   
   for t = 2 : T %% induction
      %% keep a copy %%
      alpha_bak = alpha(:,:,n);
      dpi_bak = dpi;
      daij_bak = daij;
      dbjk_bak = dbjk;
      dth_bak = dth;
      
      for c = 1 : nh
         %%% transition matrices
         cpd = struct(net.CPD{net.equiv_class(c,2)});
         %%% find out parents
         for cp = 1:nh
            %%%% alpha %%%%
            tmp(:,cp) = cpd.CPT{cp}' * alpha_bak(:,cp);
            %%%% dpi %%%%
            u = reshape(dpi_bak(:,:,:,cp),[nsh,ns1]);
            v = (u * cpd.CPT{cp})';
            tmpi(:,cp) = v(:);
            %%%% daij %%%%
            ua = reshape(daij_bak(:,:,:,:,:,cp),[nsh2,ns1]);
            va = (ua * cpd.CPT{cp})';
            tmpa(:,cp) = va(:);
            if net.discrete == 1
               %%%% dbjk %%%%
               ub = reshape(dbjk_bak(:,:,:,:,cp),[nsho,ns1]);
               vb = (ub * cpd.CPT{cp})';
               tmpb(:,cp) = vb(:);
            end
            %%%% dth %%%%
            ut = reshape(dth_bak(:,:,:,cp),[nh2,ns1]);
            vt = (ut * cpd.CPT{cp})';
            tmpt(:,cp) = vt(:);
         end
         %%% output node obs matrix
         ocpt = net.CPD{c+nh}.CPT;
         if net.discrete == 1
            x = ocpt(:,obs(c,t));
         else
            bjot(:,:,t,c) = cobs_prob(ocpt,obs(c,t));
            bjt(:,t,c) = sum(bjot(:,:,t,c),2);
            x = bjt(:,t,c);
         end
         %%% alpha %%%
         balpha = tmp * cpd.theta';
         alpha(:,c,n) = balpha .* x;
         %%% dpi %%%
         w = tmpi * cpd.theta';
         v = reshape(w,[ns1,nsh]);
         y = x(:,ones(1,nsh));
         v = v .* y;
         dpi(:,:,:,c) = reshape(v', [ns1,nh,ns1]);
         %%% daij %%%
         wa = tmpa * cpd.theta';
         va = reshape(wa, [ns1,nsh2]);
         ya = x(:,ones(1,nsh2));
         va = va .* ya;
         daij(:,:,:,:,:,c) = reshape(va', [ns1,ns1,nh,nh,ns1]);
         ijc(:,:,c) = alpha_bak(:,c) * x';
         if net.discrete == 1
            %%% dbjk %%%
            wb = tmpb * cpd.theta';
            vb = reshape(wb, [ns1,nsho]);
            yb = x(:,ones(1,nsho));
            vb = vb .* yb;
            dbjk(:,:,:,:,c) = reshape(vb', [ns1,no,nh,ns1]);
            dbjk2 = zeros(ns1,no,nh,ns1,nh);
            dbjk2(:,obs(c,t),c,:,c) = diag(balpha);
            dbjk = dbjk + dbjk2;
         end
         %%% dth %%%
         wt = tmpt * cpd.theta';
         vt = reshape(wt, [ns1,nh2]);
         yt = x(:,ones(1,nh2));
         vt = vt .* yt;
         dth(:,:,:,c) = reshape(vt', [nh,nh,ns1]);
         cj = tmp';
         cj = cj .* x(:,ones(1,nh))';
         dth2 = zeros(nh,nh,ns1,nh);
         dth2(:,c,:,c) = cj;
         dth = dth + dth2;
      end
      %%% daij (continued) %%%
      ijc2 = reshape(ijc,[ns2,nh]);
      cc = net.theta;
      tmp2 = ijc2(:) * cc(:)';
      ijcc = zeros(ns1,ns1,nh,nh);
      for c = 1 : nh
         ijcc(:,:,:,c) = reshape(tmp2(i1(c):i2(c),j1(c):j2(c)),[ns1,ns1,nh]);
      end
      daij2 = zeros(nsh3,1);
      daij2(ind) = ijcc(:);
      daij = daij + reshape(daij2,[ns1,ns1,nh,nh,ns1,nh]);
      
      if nargin==2 & n==1
         scale(t) = 1/sum(alpha(:,1,n));
      end
      alpha(:,:,n) = alpha(:,:,n) * scale(t);
      as = repmat(sum(alpha(:,:,n),1),[ns1,1]);
      gamma1(:,1,t,:) = alpha(:,:,n) ./ as;
      dpi = dpi * scale(t);
      daij = daij * scale(t);
      dbjk = dbjk * scale(t);
      dth = dth * scale(t);
   end
   %%%%% forward algorithm <-- end %%%%%
   
   %% dP/dpi %%
   pc = reshape(sum(alpha(:,:,n),1), [nh,1]);
   v = 1 ./ pc;
   u = reshape(sum(dpi,3), [nsh,nh]);
   u = u * v;
   dp = dp + reshape(u, [ns1,nh]);
   %% dP/da %%
   u = reshape(sum(daij,5), [nsh2,nh]);
   u = u * v;
   da = da + reshape(u, [ns1,ns1,nh,nh]);
   %% dP/db %%
   if net.discrete == 1
      u = reshape(sum(dbjk,4), [nsho,nh]);
      u = u * v;
      db = db + reshape(u, [ns1,no,nh]);
   else
      gamma = repmat(gamma1,[1,no,1,1]);
      bjt(bjt==0) = 1;
      gamma1(:,1,:,:) = bjt;
      if sum(gamma1(:)==0)>0, err=1; return; end
      bjot = bjot ./ repmat(gamma1,[1,no,1,1]);
      gamma = gamma .* bjot;
      db(:,:,1,:) = sum(gamma,3); %%% c_jk's
      robs11(1,1,:,:) = obs';
      robs = repmat(robs11,[ns1,no,1,1]);
      db(:,:,2,:) = sum(gamma.*robs,3); %%% u_jk's
      robs = robs - repmat(net.bjo(:,:,2,:),[1,1,T,1]);
      robs = robs .* robs;
      db(:,:,3,:) = sum(gamma.*robs,3); %%% U_jk's
      b_new = b_new + db;      
   end
   %% dP/dt %%
   u = reshape(sum(dth,3), [nh2,nh]);
   u = u * v;
   dt = dt + reshape(u, [nh,nh]);
end

if net.discrete == 0
   db1 = b_new(:,:,1,:);
   %db1(db1==0) = 1;
   if sum(db1(:)==0)>0, err=1; return; end
   b_new(:,:,2,:) = b_new(:,:,2,:) ./ db1;
   b_new(:,:,3,:) = b_new(:,:,3,:) ./ db1;
   db1 = repmat(sum(b_new(:,:,1,:),2),[1,no,1,1]);
   b_new(:,:,1,:) = b_new(:,:,1,:) ./ db1;
   db = b_new;
end

return;
