function [net, ll, err] = chmm_train(obss, NM, NH, NO, nIter, discrete, net_in, coupled)

% [net,ll,err] = chmm_train(obss,NM,NH,NO,nIter,discrete,net_in,coupled)
% Train a CHMM model using extended Baum-Welch algorithm, based on approximating
% (simplified) likelihood function. Number of forward variables is linear in the 
% number of HMMs coupled together.
% INPUT:
%    obss - observation samples (multi-dimensional array, NH by T by nobs)
%    NM - number of models coupled
%    NH - number of hidden states
%    NO - number of observation symbols
%    nIter - number of iterations
%    net_in - optional starting net
%    coupled - specifying whether or not the HMMs are coupled
% OUTPUT:
%    net - trained HMM network
%    ll - log-likelihood at each iteration
%    err - err is 0 if trained successfully, 1 otherwise

if nargin < 2, NM = 2; end
if nargin < 3, NH = 3; end
if nargin < 4, NO = 5; end
if nargin < 5, nIter = 20; end
if nargin < 6, discrete = 1; end
if nargin < 7
   net = chmm_construct(NM,NH,NO,discrete);
   if discrete == 1
      save net0 net; % save starting network
   else
      save net0c net; % save starting network
   end
else
   net = net_in;
end
if nargin < 8, coupled = 1; end

if coupled == 0
   net = decouple_all(net);
end

% initialize
[nh,T,nobs] = size(obss);
tn = nobs/2;
assert(nh==NM);

err = 0;
%ll = zeros(1,nIter);
% iteration
for i = 1 : nIter % begin of outer loop
   
   [dp,da,db,dt,alpha,scale,err] = chmm_partial(net,obss);
   if err, break; end
   pc = sum(alpha,1);
   ll(1,i) = sum(log(pc(:)))/nobs - sum(log(scale))*NM;
   if i > 1
      if ll(1,i) < ll(1,i-1)
         break;
      end
   end
   
   %disp(sprintf('i = %d, likelihoods are %10.4f', i, ll(1,i)));
   
   %% normalization
   %% pi %%
   pi = net.pi .* dp;
   tmpp = repmat(sum(pi),[NH,1]);
   if sum(tmpp(:)==0)>0, err = 1; break; end
   pi = pi ./ tmpp;
   
   %% aij %%
   aij = net.aij .* da;
   tmpa = repmat(sum(aij,2),[1,NH,1,1]);
   if coupled==0, tmpa(tmpa==0) = 1; end
   if sum(tmpa(:)==0)>0, err = 1; break; end
   aij = aij ./ tmpa;
   
   %% bjk %%
   if net.discrete == 1
      bjk = net.bjk .* db;
      tmpb = repmat(sum(bjk,2),[1,NO,1]);
      if sum(tmpb(:)==0)>0, err = 1; break; end
      bjk = bjk ./ tmpb;
   else
      bjk = db;
   end
   
   %% theta %%
   %th = net.theta;
   th = net.theta .* dt;
   tmpt = repmat(sum(th,1),[NM,1]);
   th = th ./ tmpt;
   
   % replacement
   net = update_bnet(net,pi,aij,bjk,th);
   
end; % end of outer loop

return;

